#include "stdint.h"
#include "s3c44b0x.h"
#include "delay.h"
#include "bitbang.h"
#include "uart.h"
#include "nand.h"

/* Does it really matter where we access, as long as the
   /OE or /WE lines go low?  Guess it does.  Why?
   How come ROM and SDRAM don't bother the NAND interface?
*/
#define nand_byte ((uint8_t volatile* const) 0x04000000)
#define nand_word ((uint32_t volatile* const) 0x04000000)

void nand_init()
{
   /* preset PF5 (CLE) and PF6 (ALE) low */
   *PDATF&=~(0x60);

   /* assign PF5, PF6, and PF7 to be I/O */
   *PCONF&=~(0x7fc00);
   *PCONF|=0x2400;

   /* PF7 is an input for the NAND /busy signal...assign it a pullup */
   *PUPF&=~(0x80);
}

uint32_t nand_busy()
{
   /* Nonzero--NAND device is still busy */
   /* Zero--NAND device is ready */
   return (!((*PDATF) & 0x80));
}

void nand_ale(uint32_t state)
{
   /* state=0: force NAND ALE signal low (idle)
      state=1: force NAND ALE signal high (latch in address)
   */
   if (state)
   {
      /* latch in address byte */
      *PDATF|=0x40;
   }
   else
   {
      /* idle */
      *PDATF&=~0x40;
   }
}

void nand_cle(uint32_t state)
{
   /* state=0: force NAND CLE signal low (idle)
      state=1: force NAND CLE signal high (latch in command)
   */
   if (state)
   {
      /* latch a command byte */
      *PDATF|=0x20;
   }
   else
   {
      /* idle */
      *PDATF&=~0x20;
   }
}

void nand_reset()
{
   /* issue a 0xff reset commmand to the NAND device */
   nand_init();
   nand_ale(0);
   nand_cle(0);

   nand_cle(1);
   *nand_byte=0xff;
   nand_cle(0);
}


uint8_t nand_test()
{
   /* Read the 3 byte ID code from the NAND device */
   /* return with the device size code */
   uint8_t manufacturer;
   uint8_t size_code;
   uint8_t unique_id_available;

   nand_init();
   nand_ale(0);
   nand_cle(0);

   /* command 0x90->read ID */
   nand_cle(1);
   *nand_byte=0x90;
   nand_cle(0);

   nand_ale(1);
   *nand_byte=0;
   nand_ale(0);

   /* no idea what Matrix ID is...hopefully it's 0x98! */
   manufacturer=*nand_byte;

   /* 0xe6 -> 8MB
      0x73 -> 16 MB
      0x75 -> 32 MB
      0x76 -> 64 MB
      0x79 -> 128 MB
   */
   size_code=*nand_byte;

   /* think 0xa5 means unique ID available? */
   unique_id_available=*nand_byte;

   uart_tx_string((const uint8_t*)"\r\n       Manufacturer code: 0x");
   uart_tx_string((const uint8_t*)itoh(manufacturer));
   uart_tx_string((const uint8_t*)"\r\n               Size code: 0x");
   uart_tx_string((const uint8_t*)itoh(size_code));
   uart_tx_string((const uint8_t*)"\r\n              Extra code: 0x");
   uart_tx_string((const uint8_t*)itoh(unique_id_available));

   return size_code;
}


void nand_copy(uint32_t sector, uint32_t* buffer)
{
   /* copy one sector from NAND flash to a memory buffer */
   uint32_t i;
   uint32_t shifted;

   /* disable SPI interface */
   mmc_idle();

   /* reset the NAND interface */
   nand_init();
   nand_reset();
   /* NAND command 0x00 (read 1) */
   nand_cle(1);
   nand_ale(0);
   *nand_byte=0x00;

   /* low order byte of sector address */
   nand_cle(0);
 
   nand_ale(1);
   *nand_byte=0;

   /* Hah!  This is where my bug was.
      The NAND interface doesn't want a sector address, it wants a byte address
      If we're addressing by sectors, make the first byte 0 and all is
      well.  Whew, finally!
   */

   *nand_byte=(sector & 0xff);

   /* more sector address */
   *nand_byte=((sector >> 8) & 0xff);

   /* even more sector address */
   *nand_byte=((sector >> 16) & 0xff);

   /* high byte of sector address */
   nand_ale(0);

   /* Spin until the NAND device is ready */
   /* todo: implement a timeout here */
   while (nand_busy()) {};

   /* read 512 bytes */
   for (i=0; i<128; i++)
   {
      shifted=*nand_word;
      *buffer=shifted;
      ++buffer;
   }

   /* also read the OOB area */
   for (i=0; i<4; i++)
   {
      shifted=*nand_word;
   }
}
void nand_write(uint32_t sector, uint32_t* buffer)
{
   /* write one sector from a memory buffer to  NAND flash */
   uint32_t i;
   uint32_t shifted;

   /* disable SPI interface */
  
   mmc_idle();

   /* reset the NAND interface */
   nand_init();
   nand_reset();
 
   /* NAND command 0x80 (write 1) */
   nand_cle(1);
   nand_ale(0);
   *nand_byte=0x80;
    nand_cle(0);
 
   /* low order byte of sector address */
  
   nand_ale(1);
   *nand_byte=0;

   /* Hah!  This is where my bug was.
      The NAND interface doesn't want a sector address, it wants a byte address
      If we're addressing by sectors, make the first byte 0 and all is
      well.  Whew, finally!
   */

   *nand_byte=(sector & 0xff);

   /* more sector address */
   *nand_byte=((sector >> 8) & 0xff);

   /* even more sector address */
   *nand_byte=((sector >> 16) & 0xff); 

   /* high byte of sector address */
   nand_ale(0);
  

   /* write 512 bytes */
   for (i=0; i<128; i++)
   {
     
      *nand_byte=(*buffer & 0xff);

      *nand_byte=((*buffer >> 8 ) & 0xff); 

      *nand_byte=((*buffer >> 16 ) & 0xff);

      *nand_byte=((*buffer >> 24 ) & 0xff);
     /* *nand_word=*buffer; */
      ++buffer;
   }
  
   /* also write the OOB area */
   for (i=0; i<4; i++)
   {
      *nand_word=0xFFFFFFFF;
   }
 
/* confirmation */
   nand_cle(1);
   nand_ale(0);
   *nand_byte=0x10;
    nand_cle(0);
 
   /* Spin until the NAND device is ready */
   /* todo: implement a timeout here */
   while (nand_busy()) {}; 

}

void nand_erase(uint32_t sector)
{
   /* ease  one block (32 sectors) of  NAND flashr */
   uint32_t i;
   uint32_t shifted;
    uint8_t status;

   /* disable SPI interface */
   mmc_idle();

   /* reset the NAND interface */
   nand_init();
   nand_reset();
 
   /* NAND command 0x60 (erase 1) */
   nand_cle(1);
   nand_ale(0);
   *nand_byte=0x60;

   /* low order byte of sector address */
   nand_cle(0);
  
   nand_ale(1);
/*  *nand_byte=0; */

   /* Hah!  This is where my bug was.
      The NAND interface doesn't want a sector address, it wants a byte address
      If we're addressing by sectors, make the first byte 0 and all is
      well.  Whew, finally!
   */

  *nand_byte=(sector & 0xff); 

   /* more sector address */
   *nand_byte=((sector >> 8) & 0xff);

   /* even more sector address */
/*    *nand_byte=((sector >> 16) & 0xff); */

   /* high byte of sector address */
   nand_ale(0);
 
/* confirm command */
  nand_cle(1);
  *nand_byte=0xD0;
  nand_cle(0);
 
   /* Spin until the NAND device is ready */
   /* todo: implement a timeout here */
   while (nand_busy()) { uart_tx_string((const uint8_t*)"\r\n   busy busy"); };
   delay_ms(10);
/*   nand_cle(1);
  *nand_byte=0x70;
   nand_cle(0);
   status=nand_byte; 
   nand_test(); */
}

void nand_compare(uint32_t sector, uint32_t *buffer)
{
 uint32_t temp_buffer2[256];
 uint32_t i;
 nand_copy(sector,temp_buffer2);
 for(i=0;i<128;i++) {
  if(buffer[i] != temp_buffer2[i]) {
     uart_tx_string((const uint8_t*)"\r\n   error at sector  0x");
     uart_tx_string((const uint8_t*)itoh(sector));
      uart_tx_string((const uint8_t*)"at word #");
       uart_tx_string((const uint8_t*)itoh(i));
     uart_tx_string((const uint8_t*)"\r\n org is 0x");
     uart_tx_string((const uint8_t*)itoh(buffer[i]));
     uart_tx_string((const uint8_t*)"     nand byte is 0x");
     uart_tx_string((const uint8_t*)itoh(temp_buffer2[i]));
  }
}
}
